package net.knarcraft.minecraftserverlauncher;

import net.knarcraft.minecraftserverlauncher.profile.Collection;
import net.knarcraft.minecraftserverlauncher.profile.Controller;
import net.knarcraft.minecraftserverlauncher.server.Server;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerConsoles;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerLauncherGUI;
import net.knarcraft.minecraftserverlauncher.utility.CommonFunctions;
import net.knarcraft.minecraftserverlauncher.utility.Updater;

import java.awt.*;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URISyntaxException;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import static net.knarcraft.minecraftserverlauncher.utility.CommonFunctions.stringBetween;
//Java 8 required.

/**
 * A software for managing Minecraft servers.
 *
 * @author Kristian Knarvik <kristian.knarvik@knett.no>
 * @version 1.0.0
 * @since 1.0.0
 */

public class Main {
    private static String applicationWorkDirectory;
    private static boolean serversAreRunning = false;
    private static final String updateChannel = "alpha";
    private static final String updateURL = "https://api.knarcraft.net/minecraftserverlauncher";
    private static ServerLauncherGUI gui;
    private static final Controller controller = Controller.getInstance();

    public static void main(String[] args) throws IOException {
        Updater.checkForUpdate(updateURL, updateChannel, gui);
        try (PrintWriter file = new PrintWriter(Main.getApplicationWorkDirectory() + File.separator + "latestrun.log")) {
            file.print("");
        } catch (IOException e) {
            e.printStackTrace();
        }
        EventQueue.invokeLater(() -> {
            try {
                new ServerConsoles();
                controller.loadState();
                gui = controller.getGUI();
                ScheduledExecutorService exec = Executors.newSingleThreadScheduledExecutor();
                exec.scheduleAtFixedRate(Main::updateConsoles, 10, 500, TimeUnit.MILLISECONDS);
            } catch (Exception e) {
                e.printStackTrace();
            }
        });
    }

    /**
     * Gets the controller used by the software
     *
     * @return <p>The controller used by the software</p>
     */
    public static Controller getController() {
        return controller;
    }

    /**
     * Retrieves the directory the .jar file is running from
     *
     * @return A string path
     */
    public static String getApplicationWorkDirectory() {
        if (applicationWorkDirectory == null) {
            try {
                applicationWorkDirectory = String.valueOf(new File(Main.class.getProtectionDomain().getCodeSource().getLocation().toURI().getPath()).getParentFile());
            } catch (URISyntaxException e) {
                e.printStackTrace();
                System.exit(1);
            }
        }
        return applicationWorkDirectory;
    }

    /**
     * Reads from server processes, and writes the output to consoles.
     */
    private static void updateConsoles() {
        try {
            for (Collection collection : controller.getCurrentProfile().getCollections()) {
                Server server = collection.getServer();
                if (server.isEnabled() && server.getProcess() != null) {
                    try {
                        String readText = CommonFunctions.readBufferedReader(server.getReader());
                        if (!readText.equals("")) {
                            collection.getServerConsole().output(readText);
                            updatePlayerList(readText, server);
                        }
                    } catch (IOException e) {
                        e.printStackTrace();
                    }
                    if (!server.getProcess().isAlive()) {
                        server.stopped();
                    }
                }
            }
            boolean runningNew = serversRunning();
            if (!runningNew && serversAreRunning) {
                gui.updateRunning(false);
                gui.setStatus("Servers are stopped");
            } else if (runningNew && !serversAreRunning) {
                gui.updateRunning(true);
            }
            serversAreRunning = runningNew;
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Goes through all servers and looks for any running servers.
     *
     * @return Is at least one server running?
     */
    private static boolean serversRunning() {
        int num = 0;
        for (Collection collection : controller.getCurrentProfile().getCollections()) {
            if (collection.getServer().isStarted() ||
                    (collection.getServer().getProcess() != null && collection.getServer().getProcess().isAlive())) {
                num++;
            }
        }
        return num > 0;
    }

    /**
     * Looks for strings implying a player has joined or left, and updates the appropriate lists
     *
     * @param text   <p>The text to search</p>
     * @param server <p>The server which sent the text</p>
     */
    private static void updatePlayerList(String text, Server server) {
        if (!server.getType().isProxy()) {
            String joinedPlayer = getPlayer(text, true);
            String leftPlayer = getPlayer(text, false);
            if (!joinedPlayer.equals("")) {
                if (!server.hasPlayer(joinedPlayer)) {
                    server.addPlayer(joinedPlayer);
                }
            } else if (!leftPlayer.equals("")) {
                if (server.hasPlayer(leftPlayer)) {
                    server.removePlayer(leftPlayer);
                }
            }
        }
    }

    /**
     * Searches a string for players joining or leaving
     *
     * @param text   <p>The text string to search through</p>
     * @param joined <p>Whether to search for a joining player</p>
     * @return <p>The name of a player, or an empty string</p>
     */
    private static String getPlayer(String text, boolean joined) {
        String playerName;
        if (joined) {
            playerName = stringBetween(text, "[Server thread/INFO]: ", " joined the game");
            if (playerName.equals("")) {
                playerName = stringBetween(text, "UUID of player ", " is ");
            }
        } else {
            playerName = stringBetween(text, "INFO]: ", " lost connection");
            if (playerName.equals("")) {
                playerName = stringBetween(text, "[Server thread/INFO]: ", " left the game");
            }
        }
        return playerName;
    }
}