package net.knarcraft.minecraftserverlauncher.profile;

import net.knarcraft.minecraftserverlauncher.Main;
import net.knarcraft.minecraftserverlauncher.server.Server;
import net.knarcraft.minecraftserverlauncher.server.ServerTypeHandler;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerLauncherGUI;
import net.knarcraft.minecraftserverlauncher.userinterface.ServerTab;
import net.knarcraft.minecraftserverlauncher.utility.CommonFunctions;
import net.knarcraft.minecraftserverlauncher.utility.JarDownloader;

import javax.naming.ConfigurationException;
import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.concurrent.Executors;

/**
 * This class handles profiles, GUI creation and session restoration
 */
public class Controller {

    private final String workingDirectory = Main.getApplicationWorkDirectory() + File.separator + "files";
    private final String profilesFile = workingDirectory + File.separator + "Profiles.txt";
    private final String mainFile = workingDirectory + File.separator + "Config.txt";
    private final String jarDirectory = workingDirectory + File.separator + "Jars" + File.separator;
    private final List<Profile> profileList;
    private ServerLauncherGUI serverLauncherGUI;
    private Profile currentProfile;
    private boolean downloadAllJars;
    private static Controller controller;

    /**
     * Instantiates a new controller
     */
    private Controller() {
        this.profileList = new ArrayList<>();
    }

    /**
     * Gets an instance of the controller
     *
     * @return <p>An instance of the controller</p>
     */
    public static Controller getInstance() {
        if (controller == null) {
            controller = new Controller();
        }
        return controller;
    }

    /**
     * Returns the GUI used by the software
     *
     * @return <p>The GUI used by the software</p>
     */
    public ServerLauncherGUI getGUI() {
        return this.serverLauncherGUI;
    }

    /**
     * Gets whether to download all jar files
     *
     * @return <p>Whether to download all .jar files</p>
     */
    public boolean getDownloadAllJars() {
        return downloadAllJars;
    }

    /**
     * Gets whether to run in background after closing
     *
     * @return <p>Whether to run in background after closing</p>
     */
    public boolean getRunInBackground() {
        return currentProfile.getRunInBackground();
    }

    /**
     * Gets whether to delay server startup
     *
     * @return <p>Whether to delay server startup</p>
     */
    public int getDelayStartup() {
        return currentProfile.getDelayStartup();
    }

    @Override
    public String toString() {
        int guiWidth;
        int guiHeight;
        if (serverLauncherGUI == null) {
            guiWidth = 440;
            guiHeight = 170;
        } else {
            guiWidth = serverLauncherGUI.getSize().width;
            guiHeight = serverLauncherGUI.getSize().height;
        }
        return String.format("%s;%d;%d;%b", currentProfile.getName(), guiWidth, guiHeight, downloadAllJars);
    }

    /**
     * Gets a profile by its name
     *
     * @param name <p>The name of the profile to get</p>
     * @return <p>The profile with the given name or null if not found</p>
     */
    public Profile getProfileByName(String name) {
        for (Profile profile : profileList) {
            if (profile.getName().equals(name)) {
                return profile;
            }
        }
        return null;
    }

    /**
     * Gets the names of all available profiles
     *
     * @return <p>A list of all available profiles</p>
     */
    public List<String> getProfileNames() {
        List<String> profileNames = new ArrayList<>();
        profileList.forEach((profile) -> profileNames.add(profile.getName()));
        return profileNames;
    }

    /**
     * Adds a profile if the name is valid and unique
     *
     * @param name <p>The name of the new profile</p>
     */
    public void addProfile(String name) {
        if (!CommonFunctions.nameIsValid(name)) {
            serverLauncherGUI.showError("Profile name cannot be blank and cannot contain special characters.");
            return;
        }
        for (Profile profile : profileList) {
            if (profile.getName().equals(name)) {
                serverLauncherGUI.showError("There is already a profile with this name.");
                return;
            }
        }
        Profile newProfile = new Profile(name);
        profileList.add(newProfile);
        if (currentProfile == null) {
            currentProfile = newProfile;
        }
    }

    /**
     * Removes a profile with the given name from the list of profiles, if such a profile exists
     *
     * @param name <p>The name of the profile to remove</p>
     */
    public void removeProfile(String name) {
        if (profileList.size() > 1) {
            profileList.removeIf(profile -> profile.getName().equals(name));
        } else {
            serverLauncherGUI.showError("This software requires the existence of at least one profile.");
        }
    }

    /**
     * Gets the currently selected profile
     *
     * @return <p>The currently selected profile</p>
     */
    public Profile getCurrentProfile() {
        return this.currentProfile;
    }

    /**
     * Loads data about the controller from a save string
     *
     * @param data <p>The save string to load</p>
     * @return <p>The currently selected profile</p>
     * @throws IOException <p>If unable to create the GUI</p>
     */
    private String fromString(String data) throws IOException {
        String[] dataList = data.split(";");
        this.downloadAllJars = Boolean.parseBoolean(dataList[3]);
        this.serverLauncherGUI = new ServerLauncherGUI(Integer.parseInt(dataList[1]), Integer.parseInt(dataList[2]));
        return dataList[0];
    }

    /**
     * Reads profiles and servers from a text file
     */
    public void loadState() throws ConfigurationException, IOException {
        //Loads data regarding this controller
        String currentProfile = null;
        if (new File(mainFile).exists()) {
            String controllerData = CommonFunctions.readBufferedReader(new BufferedReader(new InputStreamReader(new FileInputStream(mainFile))));
            currentProfile = this.fromString(controllerData);
        } else {
            this.serverLauncherGUI = new ServerLauncherGUI();
        }
        //Loads all saved profiles
        loadProfiles(currentProfile);
        //Makes the GUI show the loaded data
        executeGUILoadingTasks();
    }

    /**
     * Loads all saved profiles
     *
     * @param currentProfile <p>The profile saved as the current profile</p>
     * @throws ConfigurationException <p>If unable to load a profile</p>
     */
    private void loadProfiles(String currentProfile) throws ConfigurationException {
        try (Scanner in = new Scanner(new File(profilesFile))) {
            while (in.hasNextLine()) {
                String profileData = in.nextLine();
                Profile loadedProfile = Profile.fromString(profileData);
                profileList.add(loadedProfile);
            }
        } catch (FileNotFoundException e) {
            addProfile("Default");
        }

        this.currentProfile = getProfileByName(currentProfile);
        if (this.currentProfile == null) {
            this.currentProfile = this.profileList.get(0);
        }
    }

    /**
     * Updates the GUI as necessary to display loaded data
     */
    private void executeGUILoadingTasks() {
        this.serverLauncherGUI.updateProfiles();
        this.serverLauncherGUI.update();
        this.currentProfile.updateConsoles();
        if (this.downloadAllJars) {
            Executors.newSingleThreadExecutor().execute(() -> {
                JarDownloader downloader = new JarDownloader(serverLauncherGUI, jarDirectory);
                try {
                    downloader.downloadJars();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            });
        }
        if (this.currentProfile.getRunInBackground()) {
            Executors.newSingleThreadExecutor().execute(Server::startServers);
            this.serverLauncherGUI.hide();
        }
    }

    /**
     * Sets the download all jars option
     *
     * @param downloadAllJars <p>Whether to download all jars</p>
     */
    public void setDownloadAllJars(boolean downloadAllJars) {
        this.downloadAllJars = downloadAllJars;
    }

    /**
     * Sets the currently selected profile
     *
     * @param profileName <p>The name of the currently selected profile</p>
     */
    public void setCurrentProfile(String profileName) {
        this.currentProfile = getProfileByName(profileName);
    }

    /**
     * Saves the state of the entire application to disk
     */
    public void saveState() {
        saveGUIStateToServer();
        try {
            CommonFunctions.createAllFolders();
            PrintWriter mFile = new PrintWriter(mainFile);
            mFile.println(this.toString());
            mFile.close();
            PrintWriter pFile = new PrintWriter(profilesFile);
            for (Profile profile : profileList) {
                pFile.println(profile.toString());
            }
            pFile.close();
        } catch (FileNotFoundException e) {
            serverLauncherGUI.showError("Unable to save data.");
            e.printStackTrace();
        }
    }

    /**
     * Updates the server object with the current state of the GUI server tab
     */
    private void saveGUIStateToServer() {
        for (Collection collection : this.currentProfile.getCollections()) {
            Server server = collection.getServer();
            ServerTab serverTab = collection.getServerTab();
            server.setPath(serverTab.getPath());
            server.setMaxRam(serverTab.getMaxRam());
            try {
                server.setType(ServerTypeHandler.getByName(serverTab.getType()));
            } catch (ConfigurationException e) {
                e.printStackTrace();
            }
            try {
                server.setServerVersion(serverTab.getVersion());
            } catch (IllegalArgumentException e) {
                serverLauncherGUI.showError("Invalid server version for " + server.getName());
            }
            server.setEnabled(serverTab.isEnabled());
        }
    }

}
